package wpd

import (
	"bytes"
	"context"
	"errors"
	"fmt"
	"io"
	"runtime"

	"github.com/rs/zerolog"

	"orus.io/orus-io/rednerd/rendering"
)

var (
	ErrInvalidInputType  = errors.New("invalid input type")
	ErrInvalidOutputType = errors.New("invalid output type")
)

// BufferCloser is a bytes.Buffer that implements io.ReadCloser.
type BufferCloser struct {
	*bytes.Buffer
}

// Close does nothing.
func (b BufferCloser) Close() error {
	return nil
}

// Engine implements rendering.Engine.
type Engine struct {
	wpd WPD
}

// Info returns the engine description.
func (e Engine) Info() rendering.EngineInfo {
	return rendering.EngineInfo{
		Type:        "Weasyprint",
		Description: `Render HTML to PDF with weasyprint.`,
		Transitions: []rendering.EngineTransitionInfo{
			{FromType: "text/html", ToType: "application/pdf", Priority: 2.0},
		},
		Metadata: []rendering.EngineMetadataInfo{
			{
				Name:        "weasyprint-css",
				Description: `A global CSS passed to weasyprint `,
			}, {
				Name: "print-paper-size",
				Description: `A page size ("A4", "A5 landscape", "21cm 29.6cm"...).
Will be converted to CSS and combined to weasyprint-css if provided`,
			}, {
				Name: "print-paper-margin",
				Description: `
A page margin ("1cm", "1cm 5cm", "1cm 2cm 1cm 2cm", "top right bottom left"...).
Will be converted to CSS and combined to weasyprint-css if provided
			`,
			},
		},
	}
}

// Render a document from a html doc to pdf.
func (e Engine) Render(
	ctx context.Context, w rendering.DocumentWriter, r rendering.DocumentReader, toType string,
) error {
	log := zerolog.Ctx(ctx)

	if toType != "application/pdf" {
		return fmt.Errorf("%w: %s", ErrInvalidOutputType, toType)
	}

	mimetype, metadata, err := r.Metadata()
	if err != nil {
		return err
	}

	if mimetype != "text/html" {
		return fmt.Errorf("%w: %s", ErrInvalidInputType, mimetype)
	}

	css := metadata["weasyprint-css"]

	if pageSize := metadata["print-paper-size"]; pageSize != "" {
		css += fmt.Sprintf(`
@page {
	size: %s;
}
`,
			pageSize)
	}

	if pageMargin := metadata["print-paper-margin"]; pageMargin != "" {
		css += fmt.Sprintf(`
@page {
	margin: %s;
}
`,
			pageMargin)
	}

	html, err := io.ReadAll(r)
	if err != nil {
		return err
	}

	id, err := e.wpd.Submit(ctx, css, html)
	if err != nil {
		return err
	}

	done := make(chan struct{})
	defer close(done)

	go func(jobID uint64) {
		select {
		case <-ctx.Done():
			log.Warn().Uint64("job-id", jobID).Msg("Canceling wpd job")

			if err := e.wpd.Cancel(jobID); err != nil {
				log.Err(err).Uint64("job-id", jobID).Msg("could not cancel job")
			}
		case <-done:
		}
	}(id)

	rid, content, err := e.wpd.NextResult()
	if err != nil {
		if errors.Is(err, ErrWpdFailed) {
			return fmt.Errorf("%w: %w", rendering.ErrEngineCrashed, err)
		}

		return err
	}

	if rid != id {
		return fmt.Errorf("wpd: result id does not match jobid. Expected: %d, got: %d", id, rid)
	}

	w.Metadata().Update(metadata)

	if _, err := w.Write(content); err != nil {
		return err
	}

	return nil
}

//nolint:lll
type EngineOptions struct {
	WorkerPoolSize int `long:"worker-pool-size" ini-name:"worker-pool-size" env:"WORKER_POOL_SIZE" description:"Number of wpd process to run concurrently. default is the number of logical CPUs usable by the process"`
}

// NewEngine creates a wpd rendering engine.
func NewEngine(options EngineOptions, log zerolog.Logger) (*Engine, error) {
	p, err := NewRealProcess()
	if err != nil {
		return nil, fmt.Errorf("could not start wpd: %w", err)
	}

	wpd, err := NewProcessWPD(p, log)
	if err != nil {
		return nil, err
	}

	return &Engine{
		wpd: wpd,
	}, nil
}

func init() {
	rendering.RegisterRenderingEngineSetup(
		"WPD",
		func() interface{} {
			return &EngineOptions{}
		},
		func(optionGroup interface{}, log zerolog.Logger) (rendering.Engine, error) {
			options, ok := optionGroup.(*EngineOptions)
			if !ok {
				panic(fmt.Sprintf(
					"could not decode engine options from option group %#v",
					optionGroup,
				))
			}

			if options.WorkerPoolSize == 0 {
				options.WorkerPoolSize = runtime.NumCPU()
			}

			return rendering.NewPooledEngine(
				func(log zerolog.Logger) (rendering.Engine, error) {
					return NewEngine(*options, log)
				},
				options.WorkerPoolSize,
				log,
			)
		},
	)
}
