package models

import (
	"bytes"
	"encoding/base64"
	"io"
)

const (
	BodyFormatB64  = "base64"
	BodyFormatText = "text"
)

// NewDocument creates an empty Document.
func NewDocument(mimetype string) *Document {
	return &Document{
		Type: mimetype,
	}
}

// SetMetadata sets a metadata.
func (m *Document) SetMetadata(name, value string) {
	if m.Metadata == nil {
		m.Metadata = make(Metadata)
	}

	m.Metadata[name] = value
}

// UpdateMetadata sets several metadata entries at once.
func (m *Document) UpdateMetadata(entries Metadata) *Document {
	if m.Metadata == nil {
		m.Metadata = make(Metadata)
	}

	for key, value := range entries {
		m.Metadata[key] = value
	}

	return m
}

func (m *Document) ReplaceMetadata(entries Metadata) *Document {
	m.Metadata = entries

	return m
}

// SetBytes sets the document content.
func (m *Document) SetBytes(data []byte) *Document {
	m.BodyFormat = BodyFormatB64
	m.Body = base64.StdEncoding.EncodeToString(data)

	return m
}

type BytesWriter struct {
	d       *Document
	buf     bytes.Buffer
	encoder io.WriteCloser
}

func (w *BytesWriter) Write(p []byte) (int, error) {
	return w.encoder.Write(p)
}

func (w *BytesWriter) Close() error {
	err := w.encoder.Close()
	w.d.Body = w.buf.String()
	w.d.BodyFormat = BodyFormatB64

	return err
}

func (m *Document) BytesWriter() io.WriteCloser {
	w := BytesWriter{
		d: m,
	}
	w.encoder = base64.NewEncoder(base64.StdEncoding, &w.buf)

	return &w
}

// Bytes is the document content as a slice of bytes.
func (m *Document) Bytes() ([]byte, error) {
	switch m.BodyFormat {
	case "":
		fallthrough
	case BodyFormatB64:
		return base64.StdEncoding.DecodeString(m.Body)
	case BodyFormatText:
		return []byte(m.Body), nil
	default:
		return []byte(m.Body), nil
	}
}

// SetString sets the document content.
func (m *Document) SetString(data string) *Document {
	m.SetBytes([]byte(data))

	return m
}

// String is the document content as a string.
func (m *Document) String() (string, error) {
	switch m.BodyFormat {
	case "":
		fallthrough
	case BodyFormatB64:
		b, err := base64.StdEncoding.DecodeString(m.Body)
		if err != nil {
			return "", err
		}

		return string(b), nil
	case BodyFormatText:
		return m.Body, nil
	default:
		return m.Body, nil
	}
}

// AppendRenderErrors appends errors to the document.
func (m *Document) AppendRenderErrors(errors []RenderError) *Document {
	if len(errors) != 0 {
		m.RenderErrors = append(m.RenderErrors, errors...)
	}

	return m
}

// SetRenderErrors sets the 'render-errors' field.
func (m *Document) SetRenderErrors(errors []RenderError) *Document {
	if len(errors) != 0 {
		m.RenderErrors = errors
	}

	return m
}
