package handlers

import (
	"context"
	"errors"

	"github.com/Masterminds/squirrel"
	"github.com/go-openapi/runtime/middleware"
	"github.com/jmoiron/sqlx"
	"github.com/rs/zerolog"
	"golang.org/x/crypto/bcrypt"
	"orus.io/orus-io/go-orusapi/database"

	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/restapi/operations/user_management"
)

// NewUserPasswordSetHandler returns a UserPasswordSetHandler.
func NewUserPasswordSetHandler(db *sqlx.DB) *UserPasswordSetHandler {
	return &UserPasswordSetHandler{
		db: db,
	}
}

// UserPasswordSetHandler handles GET /user.
type UserPasswordSetHandler struct {
	db *sqlx.DB
}

func (h *UserPasswordSetHandler) handle(
	ctx context.Context, username string, password string,
) error {
	log := zerolog.Ctx(ctx)

	hashedPassword, err := bcrypt.GenerateFromPassword([]byte(password), 10)
	if err != nil {
		return err
	}

	q := squirrel.Update(models.UserDBTable).
		Set(models.UserDBPasswordColumn, string(hashedPassword)).
		Where(squirrel.Eq{models.UserDBUsernameColumn: username})

	tx, err := database.Begin(ctx, h.db)
	if err != nil {
		return err
	}

	defer tx.RollbackIfOpened(*log)

	r, err := database.ExecContext(ctx, tx, q, log)
	if err != nil {
		return err
	}

	count, err := r.RowsAffected()
	if err != nil {
		return err
	}

	if count == 0 {
		return ErrNoSuchUser
	}

	if err := tx.Commit(); err != nil {
		return err
	}

	return nil
}

// Handle handles the incoming request.
func (h *UserPasswordSetHandler) Handle(
	params user_management.UserPasswordSetParams, principal *models.Principal,
) middleware.Responder {
	ctx := params.HTTPRequest.Context()

	log := zerolog.Ctx(ctx)
	if !principal.IsAdmin && principal.Subject != params.Username {
		log.Error().Msg("only admins can manage user passwords, expect if you are updating your own password")

		return user_management.NewUserApikeyAddUnauthorized()
	}

	if err := h.handle(params.HTTPRequest.Context(), params.Username, params.Body.Password); err != nil {
		if errors.Is(err, ErrNoSuchUser) {
			return user_management.NewUserPasswordSetNotFound()
		}

		return InternalError(err, log, false)
	}

	return user_management.NewUserPasswordSetNoContent()
}
