package utils_test

import (
	"bytes"
	"encoding/json"
	"fmt"
	"io"
	"os"
	"runtime/debug"
	"strings"
	"testing"
	"time"

	"github.com/getsentry/sentry-go"
	"github.com/rs/zerolog"
	"github.com/stretchr/testify/require"

	"orus.io/orus-io/rednerd/utils"
)

func ExampleConsoleWriter() {
	log := zerolog.New(utils.ConsoleWriter{Out: os.Stdout, NoColor: true})

	log.Info().Str("foo", "bar").Msg("Hello World")
	// Output: <nil> INF Hello World foo=bar
}

func ExampleConsoleWriter_customFormatters() {
	out := utils.ConsoleWriter{Out: os.Stdout, NoColor: true}
	out.FormatLevel = func(i interface{}) string { return strings.ToUpper(fmt.Sprintf("%-6s|", i)) }
	out.FormatFieldName = func(i interface{}) string { return fmt.Sprintf("%s:", i) }
	out.FormatFieldValue = func(i interface{}) string { return strings.ToUpper(fmt.Sprintf("%s", i)) }
	log := zerolog.New(out)

	log.Info().Str("foo", "bar").Msg("Hello World")
	// Output: <nil> INFO  | Hello World foo:BAR
}

func ExampleNewConsoleWriter() {
	out := zerolog.NewConsoleWriter()
	out.NoColor = true // For testing purposes only
	log := zerolog.New(out)

	log.Debug().Str("foo", "bar").Msg("Hello World")
	// Output: <nil> DBG Hello World foo=bar
}

func ExampleNewConsoleWriter_customFormatters() {
	out := utils.NewConsoleWriter(
		func(w *utils.ConsoleWriter) {
			// Customize time format
			w.TimeFormat = time.RFC822
			// Customize level formatting
			w.FormatLevel = func(i interface{}) string { return strings.ToUpper(fmt.Sprintf("[%-5s]", i)) }
		},
	)
	out.NoColor = true // For testing purposes only

	log := zerolog.New(out)

	log.Info().Str("foo", "bar").Msg("Hello World")
	// Output: <nil> [INFO ] Hello World foo=bar
}

func TestConsoleLogger(t *testing.T) {
	t.Run("Numbers", func(t *testing.T) {
		buf := &bytes.Buffer{}
		log := zerolog.New(utils.ConsoleWriter{Out: buf, NoColor: true})
		log.Info().
			Float64("float", 1.23).
			Uint64("small", 123).
			Uint64("big", 1152921504606846976).
			Msg("msg")

		want := "<nil> INF msg big=1152921504606846976 float=1.23 small=123"
		if got := strings.TrimSpace(buf.String()); got != want {
			t.Errorf("\ngot:\n%s\nwant:\n%s", got, want)
		}
	})
}

func TestConsoleWriter(t *testing.T) {
	cwd, err := os.Getwd()
	require.NoError(t, err)

	exampleTime := time.Unix(0, 0).UTC().Format(time.RFC3339)

	for _, tt := range []struct {
		name       string
		w          utils.ConsoleWriter
		input      string
		expected   string
		timeFormat string
	}{
		{
			"Default field formatter",
			utils.ConsoleWriter{NoColor: true, PartsOrder: []string{"foo"}},
			`{"foo": "DEFAULT"}`,
			"DEFAULT foo=DEFAULT\n",
			zerolog.TimeFieldFormat,
		},
		{
			"Write colorized",
			utils.ConsoleWriter{NoColor: false},
			`{"level": "warn", "message": "Foobar"}`,
			"\x1b[90m<nil>\x1b[0m \x1b[31mWRN\x1b[0m Foobar\n",
			zerolog.TimeFieldFormat,
		},
		{
			"Write fields",
			utils.ConsoleWriter{NoColor: true},
			`{"time": "` + exampleTime +
				`", "level": "debug", "message": "Foobar", "foo": "bar"}`,
			"12:00AM DBG Foobar foo=bar\n",
			zerolog.TimeFieldFormat,
		},
		{
			"Unix timestamp input format",
			utils.ConsoleWriter{TimeFormat: time.StampMilli, NoColor: true},
			`{"time": 1234, "level": "debug", "message": "Foobar", "foo": "bar"}`,
			"Jan  1 00:20:34.000 DBG Foobar foo=bar\n",
			zerolog.TimeFormatUnix,
		},
		{
			"Unix timestamp ms input format",
			utils.ConsoleWriter{TimeFormat: time.StampMilli, NoColor: true},
			`{"time": 1234567, "level": "debug", "message": "Foobar", "foo": "bar"}`,
			"Jan  1 00:20:34.567 DBG Foobar foo=bar\n",
			zerolog.TimeFormatUnixMs,
		},
		{
			"Unix timestamp us input format",
			utils.ConsoleWriter{TimeFormat: time.StampMicro, NoColor: true},
			`{"time": 1234567891, "level": "debug", "message": "Foobar", "foo": "bar"}`,
			"Jan  1 00:20:34.567891 DBG Foobar foo=bar\n",
			zerolog.TimeFormatUnixMicro,
		},
		{
			"No message field",
			utils.ConsoleWriter{NoColor: true},
			`{"level": "debug", "foo": "bar"}`,
			"<nil> DBG  foo=bar\n",
			zerolog.TimeFieldFormat,
		},
		{
			"No level field",
			utils.ConsoleWriter{NoColor: true},
			`{"message": "Foobar", "foo": "bar"}`,
			"<nil> ??? Foobar foo=bar\n",
			zerolog.TimeFieldFormat,
		},
		{
			"Write colorized field",
			utils.ConsoleWriter{NoColor: false},
			`{"level": "warn", "message": "Foobar", "foo": "bar"}`,
			"\x1b[90m<nil>\x1b[0m \x1b[31mWRN\x1b[0m Foobar \x1b[36mfoo=\x1b[0mbar\n",
			zerolog.TimeFieldFormat,
		},
		{
			"Write error field",
			utils.ConsoleWriter{NoColor: true},
			`{"time": "` + exampleTime +
				`", "level": "error", "message": "Foobar", "aaa": "bbb", "error": "Error"}`,
			"12:00AM ERR Foobar error=Error aaa=bbb\n",
			zerolog.TimeFieldFormat,
		},
		{
			"Write caller field",
			utils.ConsoleWriter{NoColor: true},
			`{"time": "` + exampleTime +
				`", "level": "debug", "message": "Foobar", "foo": "bar", "caller": "` +
				cwd + `/foo/bar.go"}`,
			"12:00AM DBG foo/bar.go > Foobar foo=bar\n",
			zerolog.TimeFieldFormat,
		},
		{
			"Write json field",
			utils.ConsoleWriter{NoColor: true},
			`{"level": "debug", "message": "Foobar", "foo": [1, 2, 3], "bar": true}`,
			"<nil> DBG Foobar bar=true foo=[1,2,3]\n",
			zerolog.TimeFieldFormat,
		},
	} {
		t.Run(tt.name, func(t *testing.T) {
			if tt.timeFormat != zerolog.TimeFieldFormat {
				of := zerolog.TimeFieldFormat
				defer func() {
					zerolog.TimeFieldFormat = of
				}()

				zerolog.TimeFieldFormat = tt.timeFormat
			}

			buf := &bytes.Buffer{}
			tt.w.Out = buf

			_, err := tt.w.Write([]byte(tt.input))
			require.NoError(t, err)

			actualOutput := buf.String()

			require.Equal(t, tt.expected, actualOutput)
		})
	}

	t.Run("Write exception", func(t *testing.T) {
		buf := &bytes.Buffer{}
		w := utils.ConsoleWriter{Out: buf, NoColor: true}

		b := debug.Stack()
		t.Log(string(b))

		st := sentry.NewStacktrace()

		b, err := json.Marshal(st)
		require.NoError(t, err)

		evt := `{"level": "error", "message": "Foobar", "exception": {"type": "any", "value": "error!", "stacktrace": ` +
			string(b) + `}}`

		_, err = w.Write([]byte(evt))
		require.NoError(t, err)

		expectedOutput := fmt.Sprintf(`<nil> ERR Foobar exception=error!(any)
orus.io/orus-io/rednerd/utils_test.TestConsoleWriter.func2
    %s/log-console-writer_test.go:213

`, cwd)
		actualOutput := buf.String()
		t.Log(actualOutput)

		require.Equal(t, expectedOutput, actualOutput)
	})
}

func TestConsoleWriterConfiguration(t *testing.T) {
	t.Run("Sets TimeFormat", func(t *testing.T) {
		buf := &bytes.Buffer{}
		w := utils.ConsoleWriter{Out: buf, NoColor: true, TimeFormat: time.RFC3339}

		d := time.Unix(0, 0).UTC().Format(time.RFC3339)
		evt := `{"time": "` + d + `", "level": "info", "message": "Foobar"}`

		_, err := w.Write([]byte(evt))
		if err != nil {
			t.Errorf("Unexpected error when writing output: %s", err)
		}

		expectedOutput := "1970-01-01T00:00:00Z INF Foobar\n"
		actualOutput := buf.String()

		if actualOutput != expectedOutput {
			t.Errorf("Unexpected output:\n%q, want:\n%q", actualOutput, expectedOutput)
		}
	})

	t.Run("Sets PartsOrder", func(t *testing.T) {
		buf := &bytes.Buffer{}
		w := utils.ConsoleWriter{Out: buf, NoColor: true, PartsOrder: []string{"message", "level"}}

		evt := `{"level": "info", "message": "Foobar"}`

		_, err := w.Write([]byte(evt))
		if err != nil {
			t.Errorf("Unexpected error when writing output: %s", err)
		}

		expectedOutput := "Foobar INF\n"
		actualOutput := buf.String()

		if actualOutput != expectedOutput {
			t.Errorf("Unexpected output:\n%q, want:\n%q", actualOutput, expectedOutput)
		}
	})
}

func BenchmarkConsoleWriter(b *testing.B) {
	b.ResetTimer()
	b.ReportAllocs()

	msg := []byte(`{"level": "info", "foo": "bar", "message": "HELLO", "time": "1990-01-01"}`)

	w := utils.ConsoleWriter{Out: io.Discard, NoColor: false}

	for i := 0; i < b.N; i++ {
		_, _ = w.Write(msg)
	}
}
